' fixed point arithmetic
#define FIXED_BITS        10   ' Bit count for fractional part
#define FIXED_SCALER      (2^FIXED_BITS) ' 2 ^ FIXED_BITS
#define FIXED             integer
#define FLOAT_TO_FIXED(a) (int((a)*FIXED_SCALER))
#define INT_TO_FIXED(a)   ((a) shl FIXED_BITS)
#define FIXED_TO_INT(a)   ((a) shr FIXED_BITS)
#define FIXED_TO_FLOAT(a) (csng(a)/FIXED_SCALER)
#define MULT_FIXED(a,b)   ((a shr (FIXED_BITS/2))*(b shr (FIXED_BITS/2)))
'#define DIV_FIXED(a,b)    ((((a) shl (FIXED_BITS/2))\(b)) shl (FIXED_BITS/2))
'#define DIV_FIXED(a,b)    (((a) shl FIXED_BITS)\(b))
#define INVERSE_FIXED(a)  (DIV_FIXED(INT_TO_FIXED(1),(a)))
function DIV_FIXED(a as FIXED, b as FIXED) as FIXED
	if b=0 then return 0'&h7fffffff
	return ((a shl (FIXED_BITS/2))\b) shl (FIXED_BITS/2)
End Function

' operations:
' FIXED + FIXED
' FIXED - FIXED
' FIXED * int
' FIXED * float
' FIXED / int
' FIXED / float
' FIXED << int
' FIXED >> int
' 
' FIXED mod FIXED
' MULT_FIXED(FIXED,FIXED)
' DIV_FIXED(FIXED,FIXED)

type FColor field=1
	as ubyte R,G,B,A
end type

#define FCol(x) (*cast(FColor ptr,@(x)))

type PointF
	as single X, Y
End Type

type Trap
	as integer topY, bottomY, topX1, topX2, bottomX1, bottomX2
End Type

type Tri ' For both screen coords and UVs
	as single X1, Y1, X2, Y2, X3, Y3
End Type

type TriColor
	as integer C1, C2, C3
End Type

type TriDepth
	as single Z1, Z2, Z3
End Type

enum EStyle
	Normal
	Masked
	Translucent 'Translucent
	Modulated 'Modulated
	Multiply
	Alpha
	Warp 'Warp what was previously drawn
End Enum

' Screen buffer
dim shared as any ptr scr
dim shared as integer scr_pitch
dim shared as integer scr_w, scr_h

' Other buffers
dim shared as any ptr texbuff, colbuff

' Trafo buffer
dim shared as FIXED PerspU()
dim shared as FIXED PerspV()

' Current texture
dim shared as ubyte ptr   curtex
dim shared as integer     curtex_w, curtex_h
dim shared as integer     curtex_maxx, curtex_maxy 'w-1, h-1
dim shared as integer	  curtex_wlog
dim shared as integer ptr curtex_pal
dim shared as EStyle      curtex_style

' Current lightmap
dim shared as integer ptr lightmap
dim shared as integer	  lightmap_w, lightmap_h
dim shared as integer	  lightmap_pitchlog
dim shared as integer	  lightmap_maxx, lightmap_maxy

' Misc
dim shared as integer frames

function AddColor naked(c1 as Integer, c2 as Integer) as Integer
	asm
		movd mm0, dword ptr [esp+4]
		movd mm1, dword ptr [esp+8]
		paddusb mm0, mm1
		movd eax, mm0
		emms
		ret 8
	End Asm
End function

'v must be in 0-255
function LerpColor(c1 as integer, c2 as integer, v as integer) as Integer
	'dim as integer invertmask = &h00FF00FF
	
	asm
		'load stuff
		movd mm0, [c1]
		movd mm1, [c2]
		movd mm3, [v]
		mov eax, &h00FF00FF
		movd mm4, eax
		pxor mm5, mm5
		
		'unpack stuff
		punpcklbw mm0, mm5 'c1
		punpcklbw mm1, mm5 'c2
		punpcklwd mm3, mm3 'v
		punpckldq mm3, mm3 'v
		punpckldq mm4, mm4 'invertmask
		
		'process c2
		'c2 *= pos
		pmullw mm1, mm3
		
		'pos = 255-pos
		pxor mm3, mm4
		
		'process c1
		'c1 *= pos
		pmullw mm0, mm3
		
		psrlw mm1, 8 'c1 /= 256
		psrlw mm0, 8 'c2 /= 256
		
		'add and return
		paddusw mm0,mm1
		packuswb mm0,mm5
		movd [function],mm0
		
		emms
	End Asm
End Function

function ModColor naked(c1 as integer, c2 as integer) as integer
	asm
		movd mm0, dword ptr [esp+4]
		movd mm1, dword ptr [esp+8]
		pxor mm5, mm5
		punpcklbw mm0, mm5
		punpcklbw mm1, mm5
		pmullw mm0, mm1
		psrlw mm0, 7
		packuswb mm0, mm5
		movd eax, mm0
		emms
		ret 8
	End Asm
End Function

function MultColor naked(c1 as integer, c2 as integer) as integer
	asm
		movd mm0, dword ptr [esp+4]
		movd mm1, dword ptr [esp+8]
		pxor mm5, mm5
		punpcklbw mm0, mm5
		punpcklbw mm1, mm5
		pmullw mm0, mm1
		psrlw mm0, 8
		packuswb mm0, mm5
		movd eax, mm0
		emms
		ret 8
	End Asm
End Function


function Lerp(a as single, b as single, v as single) as Single
	'return a*(1-v)+b*v
	return (b-a)*v+a
End Function

function Min overload(a as integer, b as integer, c as integer) as integer
	dim m as integer=a
	if b<m then m=b
	if c<m then m=c
	return m
End Function

function Min(a as integer, b as integer) as integer
	if a<b then return a
	return b
End Function

function Max overload(a as integer, b as integer, c as integer) as integer
	dim m as integer=a
	if b>m then m=b
	if c>m then m=c
	return m
End Function

function Max(a as integer, b as integer) as integer
	if a>b then return a
	return b
End Function

#define SetPix(x,y,col) *cast(integer ptr,scr+(x)*4+(y)*scr_pitch)=(col)
#define GetPix(x,y) (*cast(integer ptr,scr+(x)*4+(y)*scr_pitch))

sub SafeSetPix(x as integer, y as integer, col as integer)
	if x<0 then return
	if y<0 then return
	if x>=scr_w then return
	if y>=scr_h then return
	
	SetPix(x,y,col)
End Sub

' Rounding: Grid must be power of two.

function RoundDown naked(num as integer, grid as integer) as Integer
	'return num and not(grid-1)
	asm
		mov eax, dword ptr [esp+8]
		dec eax
		not eax
		and eax, dword ptr [esp+4]
		ret 8
	End Asm
End Function

function RoundUp naked(num as integer, grid as integer) as Integer
	'return ((num-1) and not(grid-1))+grid
	asm
		mov eax, dword ptr [esp+4]
		dec eax
		mov ecx, dword ptr [esp+8]
		mov ebx, ecx
		dec ebx
		not ebx
		and eax, ebx
		add eax, ecx
		ret 8
	End Asm
End Function

sub BlurEffect()
	dim as integer x, y
	
	for y=1 to scr_h-1
		asm
			'ecx - scr_pitch
			mov ecx, [scr_pitch]
			
			'eax - dstptr
			mov eax, [y]
			mul ecx
			add eax, 4
			add eax, [scr]
			
			'ebx - ptrmax
			mov ebx, eax
			sub ebx, 8
			add ebx, ecx
			
			beg:
				mov edx, [eax] '(x,y)
				mov esi, [eax-4] '(x-1,y)
				mov edi, eax '(x,y-1)
				sub edi, ecx
				mov edi, [edi]
				
				and edx, &hFCFCFC
				and esi, &hFCFCFC
				and edi, &hFCFCFC
				
				shr edx, 2
				shr esi, 2
				shr edi, 2
				
				movd mm0, edx
				movd mm1, esi
				movd mm2, edi
				
				paddusb mm0, mm1
				paddusb mm0, mm2
				
				movd [eax], mm0
				
				add eax, 4
				cmp eax, ebx
			jle beg
			
			emms
		End Asm
	next
	
	for y=scr_h-2 to 0 step -1
		asm
			'ecx - scr_pitch
			mov ecx, [scr_pitch]
			
			'eax - ptrmax
			mov eax, [y]
			mul ecx
			add eax, [scr]
			
			'ebx - dstptr
			mov ebx, eax
			add ebx, ecx
			sub ebx, 8
			
			beg1:
				mov edx, [ebx] '(x,y)
				mov esi, [ebx+4] '(x+1,y)
				mov edi, [ebx+ecx] '(x,y+1)
				
				and esi, &hFCFCFC
				and edi, &hFCFCFC
				
				shr esi, 2
				shr edi, 2
				
				movd mm0, edx
				movd mm1, esi
				movd mm2, edi
				
				paddusb mm0, mm1
				paddusb mm0, mm2
				
				movd [ebx], mm0
				
				sub ebx, 4
				cmp ebx, eax
			jge beg1
			
			emms
		End Asm
	next
End Sub

sub WaveEffect()
	dim as integer x, y
	const amp=10
	
	static as boolean wasInit
	const tableSize=256
	const twoPi=2*3.1415926536
	static as single sinTable(tableSize-1)
	
	if wasInit=false then
		for i as integer=0 to tableSize-1
			sinTable(i) = sin(i/tableSize*twoPi)
		Next
		wasInit = true
	EndIf
	
	
	for y=0 to scr_h-1-amp
		for x=0 to scr_w-1-amp
			SetPix(x,y,GetPix(cint(x+amp+sinTable(((y+frames)/10/twoPi*tableSize) and (tableSize-1))*amp/4),cint(y+amp+sinTable(((x+frames)/10/twoPi*tableSize) and (tableSize-1))*amp/4)))
		Next
	Next
End Sub

public sub Init(buffer as any ptr, w as integer, h as integer, pitch as integer) export
	scr = buffer
	scr_w = w
	scr_h = h
	scr_pitch = pitch
	
	if texbuff <> 0 then	deallocate(texbuff)
	if colbuff <> 0 then	deallocate(colbuff)
	
	texbuff = allocate(scr_h*scr_pitch)
	colbuff = allocate(scr_h*scr_pitch)
	
	redim PerspU(RoundUp(w,16)/16,h)
	redim PerspV(RoundUp(w,16)/16,h)
	
End Sub

sub ClearBuffer(buff as any ptr, col as integer)
	dim as integer x, y
	
	for y=0 to scr_h-1
		for x=0 to scr_w-1
			*cast(integer ptr, buff+y*scr_pitch+x*4) = col
		Next
	Next
End Sub

sub DrawLine(x1 as integer, y1 as integer, x2 as integer, y2 as integer, col as integer)
	dim points as integer
	dim perc as single
	if abs(x1-x2)>abs(y1-y2) then
		points = abs(x1-x2)
	else
		points = abs(y1-y2)
	EndIf
	
	dim i as integer
	for i=0 to points-1
		perc = i/(points-1)
		SafeSetPix(lerp(x1,x2,perc), lerp(y1,y2,perc), col)
	Next
End Sub

sub FlashAdd(col as integer) export
	asm
		mov ebx, [scr]
		mov eax, [scr_pitch]
		mul dword ptr [scr_h]
		sub eax, 8
		add eax, [scr]
		movd mm1, [col]
		punpckldq mm1, mm1
		
		flashbeg:
			movq mm0, [ebx]
			paddusb mm0, mm1
			movq [ebx], mm0
			
			add ebx, 8
			cmp ebx, eax
		jle flashbeg
		
		emms
	End Asm
End Sub

sub FlashSub(col as integer) export
	asm
		mov ebx, [scr]
		mov eax, [scr_pitch]
		mul dword ptr [scr_h]
		sub eax, 8
		add eax, [scr]
		movd mm1, [col]
		punpckldq mm1, mm1
		
		flashbeg1:
			movq mm0, [ebx]
			psubusb mm0, mm1
			movq [ebx], mm0
			
			add ebx, 8
			cmp ebx, eax
		jle flashbeg1
		
		emms
	End Asm
End Sub

sub FlashMod(col as integer) export
	asm
		mov ebx, [scr]
		mov eax, [scr_pitch]
		mul dword ptr [scr_h]
		sub eax, 4
		add eax, [scr]
		
		flashbeg2:
			movd mm0, [ebx]
			movd mm1, [col]
			pxor mm5, mm5
			punpcklbw mm0, mm5
			punpcklbw mm1, mm5
			pmullw mm0, mm1
			psrlw mm0, 7
			packuswb mm0, mm5
			movd [ebx], mm0
			
			add ebx, 4
			cmp ebx, eax
		jle flashbeg2
		
		emms
	End Asm
End Sub

sub BeginUpdate export
	'ClearBuffer(scr,rgb(255,0,255))
	'ClearBuffer(texbuff,rgb(255,0,255))
End Sub

sub EndUpdate export
	'WaveEffect()
	'FlashMod(rgb(0,128,255))
	
	frames += 1
End Sub

sub FixTrap(byref tr as Trap)
	if tr.topY>tr.bottomY then
		swap tr.topY, tr.bottomY
		swap tr.topX1, tr.bottomX1
		swap tr.topX2, tr.bottomX2
	EndIf
	if tr.topX1>tr.topX2 then swap tr.topX1, tr.topX2
	if tr.bottomX1>tr.bottomX2 then swap tr.bottomX1, tr.bottomX2
End Sub

sub BuffCopy(dst as any ptr, src as any ptr, byref region as Trap, blending as EStyle)
	dim as integer x, y, xmin, xmax
	dim as integer ptr curdst, cursrc, ptrmax, roundedptrmax
	dim as FIXED xminf, xmaxf, xminstep, xmaxstep
	
	if region.topY=region.bottomY then return
	
	FixTrap(region)
	
	xminf = FLOAT_TO_FIXED(region.topX1)
	xmaxf = FLOAT_TO_FIXED(region.topX2)
	
	xminstep = FLOAT_TO_FIXED((region.bottomX1-region.topX1)/(region.bottomY-region.topY))
	xmaxstep = FLOAT_TO_FIXED((region.bottomX2-region.topX2)/(region.bottomY-region.topY))
	
	for y=region.topY to region.bottomY-1
		xmin = FIXED_TO_INT(xminf)
		xmax = FIXED_TO_INT(xmaxf)
		
		curdst = dst+scr_pitch*y+xmin shl 2
		cursrc = src+scr_pitch*y+xmin shl 2
		
		ptrmax = dst+scr_pitch*y+(xmax-1) shl 2
		'Basically, the same as ptrmax, but makes sure the length (in pixels) is even.
		roundedptrmax = dst+scr_pitch*y+(xmin+RoundDown(xmax-xmin,2)-1) shl 2
		
		select case as const blending
			case EStyle.Normal, EStyle.Masked
				asm
					mov esi, [cursrc]
					mov edi, [curdst]
					mov eax, [roundedptrmax]
					
					copybegin:
						movq mm0, [esi]
						movq [edi], mm0
						
						add esi, 8
						add edi, 8
						
						cmp edi, eax
					jle copybegin
					
					' Remainder
					cmp eax, [ptrmax]
					je waseven
						mov ebx, [esi]
						mov [edi], ebx
					waseven:
					
					emms
					
				End Asm
			case EStyle.Translucent
				asm
					mov esi, [cursrc]
					mov edi, [curdst]
					mov eax, [roundedptrmax]
					
					copybegin1:
						movq mm0, [esi]
						paddusb mm0, [edi]
						movq [edi], mm0
						
						add esi, 8
						add edi, 8
						
						cmp edi, eax
					jle copybegin1
					
					' Remainder
					cmp eax, [ptrmax]
					je waseven1
						movd mm0, [esi]
						paddusb mm0, [edi]
						movd [edi], mm0
					waseven1:
					
					emms
					
				End Asm
			case EStyle.Modulated
				asm
					mov esi, [cursrc]
					mov edi, [curdst]
					mov eax, [ptrmax]
					
					copybegin2:
						'mm0 = esi*edi
						movd mm0,[esi]
						movd mm1,[edi]
						pxor mm5,mm5
						punpcklbw mm0,mm5
						punpcklbw mm1,mm5
						pmullw mm0,mm1
						psrlw mm0, 7
						packuswb mm0,mm5
						'mm0 *= 2
						paddusb mm0, mm0
						
						movd [edi], mm0
						
						add esi, 4
						add edi, 4
						
						cmp edi, eax
					jle copybegin2
					
					emms
					
				End Asm
			case EStyle.Multiply
				asm
					mov esi, [cursrc]
					mov edi, [curdst]
					mov eax, [ptrmax]
					
					copybegin3:
						'mm0 = esi*edi
						movd mm0,[esi]
						movd mm1,[edi]
						pxor mm5,mm5
						punpcklbw mm0,mm5
						punpcklbw mm1,mm5
						pmullw mm0,mm1
						psrlw mm0, 8
						packuswb mm0,mm5
						'mm0 *= 2
						paddusb mm0, mm0
						
						movd [edi], mm0
						
						add esi, 4
						add edi, 4
						
						cmp edi, eax
					jle copybegin3
					
					emms
					
				End Asm
			case EStyle.Warp
				dim as integer warpx, warpy
				
				for x=xmin to xmax-1
					warpx = x+(*cursrc and 255) shr 2
					warpy = y+((*cursrc shr 8) and 255) shr 2
					warpx = warpx mod scr_w
					warpy = warpy mod scr_h
					
					*curdst = *cast(integer ptr,dst+warpy*scr_pitch+warpx*4)
					
					cursrc += 1
					curdst += 1
				Next
				
		End Select
		
		xminf += xminstep
		xmaxf += xmaxstep
	Next
End Sub

public sub Unload export
	if texbuff <> 0 then deallocate(texbuff)
	if colbuff <> 0 then deallocate(colbuff)
End Sub

public sub Pause(ms as integer) export
	sleep ms,1
End Sub

public sub SetTexture(tex as ubyte ptr, tex_w as integer, tex_h as integer, tex_pal as integer ptr, tex_style as EStyle) export
	curtex = tex
	curtex_w = tex_w
	curtex_h = tex_h
	curtex_maxx = curtex_w-1
	curtex_maxy = curtex_h-1
	asm
		bsf eax, [curtex_w]
		mov [curtex_wlog], eax
	End Asm
	curtex_pal = tex_pal
	curtex_style = tex_style
End Sub

public sub SetLightmap(litmap as integer ptr, litmap_w as integer, litmap_h as integer) export
	lightmap = litmap
	lightmap_w = litmap_w
	lightmap_h = litmap_h
	lightmap_maxx = lightmap_w-1
	lightmap_maxy = lightmap_h-1
	asm
		bsf eax, [lightmap_w]
		add eax, 2
		mov [lightmap_pitchlog], eax
	End Asm
End Sub

'#define FColorToFB(col) ((col).B or ((col).G shl 8) or ((col).R shl 16))

function FColorToFB naked(col as integer) as integer
	asm
		mov eax, dword ptr [esp+4]
		bswap eax
		shr eax, 8
		ret 4
	End Asm
End Function

function GetTexPix(x as integer, y as integer) as integer
	asm
		'get coords
		mov ebx, [x]
		mov eax, [y]
		and ebx, [curtex_maxx]
		and eax, [curtex_maxy]
		
		'calc texel pointer
		mov cl, [curtex_wlog]
		shl eax, cl 'a *= curtex_w
		add eax, ebx 'a += x
		add eax, [curtex] 'a += curtex
		
		'read texel value
		movzx eax, byte ptr [eax]
		
		'calc palette pointer
		shl eax, 2
		add eax, [curtex_pal]
		
		'read color
		mov eax, [eax]
		
		'bgr->rgb
		bswap eax
		shr eax, 8
		
		mov [function], eax
	End Asm
End Function

' TODO: convert this to asm
function GetLightmapPix(x as integer, y as integer) as Integer
	if lightmap=0 then return &hFFFFFF
	
	x = x and (lightmap_w-1)
	y = y and (lightmap_h-1)
	
	return lightmap[y*lightmap_w+x]
End Function

public sub DrawTile(x as single, y as single, xl as single, yl as single, _
						  u as single, v as single, ul as single, vl as single, _
						  uclamp as single, vclamp as single, col as integer) export
	
	'current UV in pixels
	dim as single cur_u,cur_v,ustep,vstep
	dim as integer cur_u_int,cur_v_int
	
	dim as integer ix,iy,xmin,xmax,ymin,ymax,curpix
	dim as any ptr dstaddr
	
	'palette index of current pixel
	dim as ubyte pind
	dim as integer pcol
	
	xmin = x
	xmax = x+xl
	ymin = y
	ymax = y+yl
	
	/'if xmin<0 then
		u -= xmin*xl
		xmin=0
	EndIf
	if xmax>=scr_w then xmax=scr_w-1
	if ymin<0 then
		v -= ymin*yl
		ymin=0
	EndIf
	if ymax>=scr_h then ymax=scr_h-1
	'/
	
	cur_u = u*ul/xl
	cur_v = v*vl/yl
	
	ustep = ul/xl
	vstep = vl/yl
	
	for iy=ymin to ymax
		cur_u = (xmin-x+u)*ul/xl
		
		for ix=xmin to xmax
			cur_u_int = cur_u
			cur_v_int = cur_v
			
			cur_u_int = cur_u_int and curtex_maxx
			cur_v_int = cur_v_int and curtex_maxy
			
			pind = curtex[cur_u_int+cur_v_int*curtex_w]
			pcol = curtex_pal[pind]
			
			curpix = MultColor(FColorToFB(pcol),col)
			
			dstaddr = scr+ix*4+iy*scr_pitch
			
			select case as const curtex_style
				case Normal
					*cast(integer ptr,dstaddr) = curpix
				case Masked
					if pind<>0 then
						*cast(integer ptr,dstaddr) = curpix
					EndIf
				case Translucent
					*cast(integer ptr,dstaddr) = AddColor(*cast(integer ptr,dstaddr), curpix)
				case Modulated
					*cast(integer ptr,dstaddr) = ModColor(*cast(integer ptr,dstaddr),curpix)
			End Select
			
			cur_u += ustep
		Next
		
		cur_v += vstep
	Next
	
End Sub

'================================
' 	Triangle drawing
'================================

type HalfTri
	as integer peakX, peakY, horX1, horX2, horY
End Type

type HalfTriUV
	as single peakU, peakV, horX1U, horX1V, horX2U, horX2V
End Type

type HalfTriColor
	as integer peak, horX1, horX2
End Type

function HalfTriToTrap(byref ht as HalfTri) as Trap
	dim t as Trap
	t.topY = ht.peakY
	t.bottomY = ht.horY
	t.topX1 = ht.peakX
	t.topX2 = ht.peakX
	t.bottomX1 = ht.horX1
	t.bottomX2 = ht.horX2
	return t
End Function

sub DrawAffineHalfColor(tr as HalfTri, col as HalfTriColor)
	dim as single r, g, b
	dim as single rl, gl, bl
	dim as single rr, gr, br
	dim as single rlstep, glstep, blstep
	dim as single rrstep, grstep, brstep
	dim as single rstep, gstep, bstep
	
	dim ydir as integer
	
	if tr.peakY>tr.horY then
		ydir=-1
		'don't go outside screen (Y)
		if tr.peakY>=scr_h then
			tr.peakY = scr_h-1
		EndIf
	else
		ydir=1
		'don't go outside screen (Y)
		if tr.horY>=scr_h then
			tr.horY = scr_h-1
		EndIf
	EndIf
	
	if tr.horX2<tr.horX1 then
		swap tr.horX2,tr.horX1
		swap col.horX2,col.horX1
	EndIf
	
	dim height as integer = abs(tr.horY-tr.peakY)+1
	
	rl = FCol(col.peak).R
	gl = FCol(col.peak).G
	bl = FCol(col.peak).B
	
	rr = FCol(col.peak).R
	gr = FCol(col.peak).G
	br = FCol(col.peak).B
	
	rlstep = (FCol(col.horX1).R-FCol(col.peak).R)/height
	glstep = (FCol(col.horX1).G-FCol(col.peak).G)/height
	blstep = (FCol(col.horX1).B-FCol(col.peak).B)/height
	
	rrstep = (FCol(col.horX2).R-FCol(col.peak).R)/height
	grstep = (FCol(col.horX2).G-FCol(col.peak).G)/height
	brstep = (FCol(col.horX2).B-FCol(col.peak).B)/height
	
	dim y as integer
	for y=tr.peakY to tr.horY step ydir
		
		dim as single perc
		dim as integer xmin, xmax
		
		perc = (y-tr.peakY)/height*ydir
		
		xmin = lerp(tr.peakX,tr.horX1,perc)
		xmax = lerp(tr.peakX,tr.horX2,perc)
		
		if xmin<0 then xmin=0
		if xmax>=scr_w then xmax=scr_w-1
		
		r = rl
		g = gl
		b = bl
		
		rstep = (rr-r)/(xmax-xmin+1)
		gstep = (gr-g)/(xmax-xmin+1)
		bstep = (br-b)/(xmax-xmin+1)
		
		dim x as Integer
		for x=xmin to xmax
			*cast(integer ptr, colbuff+scr_pitch*y+4*x) = rgb(r,g,b)
			
			r += rstep
			g += gstep
			b += bstep
		Next
		
		rl += rlstep
		gl += glstep
		bl += blstep
		
		rr += rrstep
		gr += grstep
		br += brstep
	Next
	
End Sub

sub DrawAffineHalfTex(tr as HalfTri, uv as HalfTriUV)
	static as integer i,j,ydir,xmin,xmax
	static as single ypc 'y percent (0..1)
	static as any ptr dstptr, ptrmax
	
	if tr.peakY>tr.horY then
		ydir=-1
		'don't go outside screen (Y)
		if tr.peakY>=scr_h then
			tr.peakY = scr_h-1
		EndIf
	else
		ydir=1
		'don't go outside screen (Y)
		if tr.horY>=scr_h then
			tr.horY = scr_h-1
		EndIf
	EndIf
	
	if tr.horX2<tr.horX1 then
		swap tr.horX2,tr.horX1
		swap uv.horX2U,uv.horX1U
		swap uv.horX2V,uv.horX1V
	EndIf
	
	'used in interpolation
	static as FIXED ucur, vcur, ustep, vstep
	static as FIXED ucurfix, vcurfix, ustepfix, vstepfix
	static as integer xlen
	static as integer backup_esp, backup_ebp
	static as FIXED ul, vl, ur, vr, ulystep, vlystep, urystep, vrystep
	
	static as single inv_height
	inv_height=1/abs(tr.horY-tr.peakY)
	
	ulystep = FLOAT_TO_FIXED((uv.horX1U-uv.peakU)*inv_height)
	vlystep = FLOAT_TO_FIXED((uv.horX1V-uv.peakV)*inv_height)
	urystep = FLOAT_TO_FIXED((uv.horX2U-uv.peakU)*inv_height)
	vrystep = FLOAT_TO_FIXED((uv.horX2V-uv.peakV)*inv_height)
	
	ul = FLOAT_TO_FIXED(uv.peakU)
	vl = FLOAT_TO_FIXED(uv.peakV)
	ur = FLOAT_TO_FIXED(uv.peakU)
	vr = FLOAT_TO_FIXED(uv.peakV)
	
	static as single xminf,xmaxf
	static as single xminstep,xmaxstep
	
	xminf=tr.peakX
	xmaxf=tr.peakX
	
	xminstep = (tr.horX1-tr.peakX)*inv_height
	xmaxstep = (tr.horX2-tr.peakX)*inv_height
	
	
	
	for i=tr.peakY to tr.horY step ydir
		'don't go outside screen (Y)
		'if i<0 or i>=cur_h then continue for
		'if i>=cur_h then continue for
		
		xmin=xminf
		xmax=xmaxf
		
		'if xmin=xmax then goto continueFor
		
		'ustepfix = (ur-ul)\(xmax-xmin)
		'vstepfix = (vr-vl)\(xmax-xmin)
		
		asm
			mov eax, [i]
			mul dword ptr [scr_pitch]
			mov ebx, [xmin]
			
			shl ebx, 2
			add eax, ebx
			
			add eax, [texbuff]
			mov [dstptr], eax
			
			
			mov ebx, [xmax]
			sub ebx, [xmin]
			inc ebx
			shl ebx, 2
			add ebx, eax
			mov [ptrmax], ebx
			
			
			mov eax, [ul]
			mov ebx, [vl]
			mov [ucurfix], eax
			mov [vcurfix], ebx
			
			mov edi, [dstptr]
			
			'calc u/vstepfix
			'allowed to use eax, ebx, ecx, edx, esi
			
			mov ecx, [xmax]
			sub ecx, [xmin]
			cmp ecx, 0
			je skipLine 'don't divide by zero
			
			mov eax, [ur]
			sub eax, [ul]
			cdq
			idiv ecx
			mov [ustepfix], eax
			
			mov eax, [vr]
			sub eax, [vl]
			cdq
			idiv ecx
			mov [vstepfix], eax
			
			
			mov [backup_esp], esp
			mov [backup_ebp], ebp
			
			mov esp, [ucurfix]
			mov ebp, [vcurfix]
			
			mov edx, [ustepfix]
			mov esi, [vstepfix]
			
			mov cl, [curtex_wlog]
			
			start:
				' eax, ebx - for many things
				' cl  - curtex_wlog
				' edx - ustepf
				' esi - vstepf
				' edi - dstptr
				' esp - ucurf
				' ebp - vcurf
				
				'take u,v coords
				mov ebx, esp
				mov eax, ebp
				'convert from fixed point to int
				shr ebx, FIXED_BITS
				shr eax, FIXED_BITS
				'wrap
				and ebx, [curtex_maxx]
				and eax, [curtex_maxy]
				'calc texel pointer
				shl eax, cl
				add eax, ebx
				add eax, [curtex]
				'read texel value
				movzx eax, byte ptr [eax]
				'calc palette pointer
				shl eax, 2
				add eax, [curtex_pal]
				'read color
				mov eax, [eax]
				'bgr->rgb
				bswap eax
				shr eax, 8
				'draw the pixel
				mov [edi], eax
				
				add edi, 4
				add esp, edx
				add ebp, esi
				
				cmp edi, [ptrmax]
			js start
			
			mov esp, [backup_esp]
			mov ebp, [backup_ebp]
			
			skipLine:
			
			mov eax, [ul]
			mov ebx, [vl]
			mov ecx, [ur]
			mov edx, [vr]
			add eax, [ulystep]
			add ebx, [vlystep]
			add ecx, [urystep]
			add edx, [vrystep]
			mov [ul], eax
			mov [vl], ebx
			mov [ur], ecx
			mov [vr], edx
		End Asm
		
		'continueFor:
		
		xminf += xminstep
		xmaxf += xmaxstep
	Next
End Sub

function LineXFromY(x1 as single, y1 as single, x2 as single, y2 as single, y as single) as single
	if y1=y2 then return x1
	
	return ( (x2-x1)*(y-y1)+x1*(y2-y1) ) / ( y2-y1 )
	
End Function

sub SplitTri(tri as Tri, byref top as Trap, byref bottom as Trap)
	'find point to split along
	
	'find proper top/middle/bottom points
	'v - middle point to split along
	'x - new point made after splitting
	dim as single topX,topY,vX,vY,xX,botX,botY
	dim as ubyte midv
	if (tri.Y2<=tri.Y1 and tri.Y1<=tri.Y3) or (tri.Y3<=tri.Y1 and tri.Y1<=tri.Y2) then
		vX=tri.X1
		vY=tri.Y1
		topX=tri.X3
		topY=tri.Y3
		botX=tri.X2
		botY=tri.Y2
	elseif (tri.Y1<=tri.Y2 and tri.Y2<=tri.Y3) or (tri.Y3<=tri.Y2 and tri.Y2<=tri.Y1) then
		vX=tri.X2
		vY=tri.Y2
		topX=tri.X1
		topY=tri.Y1
		botX=tri.X3
		botY=tri.Y3
	else
		vX=tri.X3
		vY=tri.Y3
		topX=tri.X2
		topY=tri.Y2
		botX=tri.X1
		botY=tri.Y1
	EndIf
	
	'calculate X coordinate of split point
	xX = LineXFromY(topX,topY,botX,botY,vY)
	
	top.topY = topY
	top.topX1 = topX
	top.topX2 = topX
	top.bottomY = vY
	top.bottomX1 = vX
	top.bottomX2 = xX
	
	bottom.topY = vY
	bottom.topX1 = vX
	bottom.topX2 = xX
	bottom.bottomY = botY
	bottom.bottomX1 = botX
	bottom.bottomX2 = botX
	
	/'dim as HalfTri      htt, htb
	
	htt.peakX=topX
	htt.peakY=topY
	htt.horX1=vX
	htt.horX2=xX
	htt.horY=vY
	
	htb.peakX=botX
	htb.peakY=botY
	htb.horX1=vX
	htb.horX2=xX
	htb.horY=vY
	
	top    = HalfTriToTrap(htt)
	bottom = HalfTriToTrap(htt)'/
End sub

sub DrawTri(tr as Tri, uv as Tri, col as TriColor) export
	
	'Top and bottom can in fact be on opposite sides - it
	'doesn't matter as DrawTriHalf doesn't care about order.
	
	dim as single  topX,topY,vX,vY,xX,botX,botY
	dim as single  topU,botU,vU,xU
	dim as single  topV,botV,vV,xV
	dim as integer topC,botC,vC,xC
	
	'find point to split along
	
	'find proper top/middle/bottom points
	'v - middle point to split along
	'x - new point made after splitting
	if (tr.Y2<=tr.Y1 and tr.Y1<=tr.Y3) or (tr.Y3<=tr.Y1 and tr.Y1<=tr.Y2) then
		vX=tr.X1
		vY=tr.Y1
		vU=uv.X1
		vV=uv.Y1
		vC=col.C1
		
		topX=tr.X3
		topY=tr.Y3
		topU=uv.X3
		topV=uv.Y3
		topC=col.C3
		
		botX=tr.X2
		botY=tr.Y2
		botU=uv.X2
		botV=uv.Y2
		botC=col.C2
	elseif (tr.Y1<=tr.Y2 and tr.Y2<=tr.Y3) or (tr.Y3<=tr.Y2 and tr.Y2<=tr.Y1) then
		vX=tr.X2
		vY=tr.Y2
		vU=uv.X2
		vV=uv.Y2
		vC=col.C2
		
		topX=tr.X1
		topY=tr.Y1
		topU=uv.X1
		topV=uv.Y1
		topC=col.C1
		
		botX=tr.X3
		botY=tr.Y3
		botU=uv.X3
		botV=uv.Y3
		botC=col.C3
	else
		vX=tr.X3
		vY=tr.Y3
		vU=uv.X3
		vV=uv.Y3
		vC=col.C3
		
		topX=tr.X2
		topY=tr.Y2
		topU=uv.X2
		topV=uv.Y2
		topC=col.C2
		
		botX=tr.X1
		botY=tr.Y1
		botU=uv.X1
		botV=uv.Y1
		botC=col.C1
	EndIf
	
	
	'calculate the x point
	xX = LineXFromY(topX,topY,botX,botY,vY)
	
	dim as single xPc=(vY-topY)/(botY-topY)
	xU = lerp(topU,botU,xPc)
	xV = lerp(topV,botV,xPc)
	xC = LerpColor(topC,botC,xPc*255)
	
	dim as HalfTri      htt, htb
	dim as HalfTriUV    hut, hub
	dim as HalfTriColor hct, hcb
	
	htt.peakX=topX
	htt.peakY=topY
	htt.horX1=vX
	htt.horX2=xX
	htt.horY=vY
	
	hut.peakU=topU
	hut.peakV=topV
	hut.horX1U=vU
	hut.horX1V=vV
	hut.horX2U=xU
	hut.horX2V=xV
	
	hct.peak=topC
	hct.horX1=vC
	hct.horX2=xC
	
	htb.peakX=botX
	htb.peakY=botY
	htb.horX1=vX
	htb.horX2=xX
	htb.horY=vY
	
	hub.peakU=botU
	hub.peakV=botV
	hub.horX1U=vU
	hub.horX1V=vV
	hub.horX2U=xU
	hub.horX2V=xV
	
	hcb.peak=botC
	hcb.horX1=vC
	hcb.horX2=xC
	
	DrawAffineHalfTex  (htt,hut)
	DrawAffineHalfTex  (htb,hub)
	
	DrawAffineHalfColor(htt,hct)
	DrawAffineHalfColor(htb,hcb)
	
	BuffCopy(texbuff,colbuff,HalfTriToTrap(htt),Multiply)
	BuffCopy(texbuff,colbuff,HalfTriToTrap(htb),Multiply)
	
	BuffCopy(scr,texbuff,HalfTriToTrap(htt),curtex_style)
	BuffCopy(scr,texbuff,HalfTriToTrap(htb),curtex_style)
End Sub

'==========================================
' Perspective corrected triangle drawing
'==========================================

#macro DrawTexTrap_Iter
	'get uv
	mov ebx, esi
	mov eax, edi
	shr ebx, FIXED_BITS
	shr eax, FIXED_BITS
	and ebx, [curtex_maxx]
	and eax, [curtex_maxy]
	
	'calc texel pointer
	mov cl, [curtex_wlog]
	shl eax, cl 'a *= curtex_w
	add eax, ebx 'a += x
	add eax, [curtex] 'a += curtex
	
	'read texel value
	movzx eax, byte ptr [eax]
	
	'calc palette pointer
	shl eax, 2
	add eax, [curtex_pal]
	
	'read color
	mov eax, [eax]
	
	'bgr->rgb
	bswap eax
	shr eax, 8
	
	'draw pixel
	mov [edx], eax
	
	'move pointer
	add edx, 4
	
	'update uv
	add esi, esp
	add edi, ebp
#EndMacro

sub DrawTexTrap(tr as Trap)
	static as integer backup_esp, backup_ebp
	
	dim as integer i,j,k,fastmin,fastmax,mincell,maxcell
	dim as FIXED xmin,xmax,xminstep,xmaxstep
	
	dim as any ptr dstptr
	
	FixTrap(tr)
	
	if tr.bottomY=tr.topY then	return
	if tr.bottomY >= scr_h then tr.bottomY = scr_h-1
	if tr.topY < 0 then tr.topY = 0
	
	dim as FIXED u,v,ustep,vstep
	
	xminstep = FLOAT_TO_FIXED((tr.bottomX1-tr.topX1)/(tr.bottomY-tr.topY))
	xmaxstep = FLOAT_TO_FIXED((tr.bottomX2-tr.topX2)/(tr.bottomY-tr.topY))
	
	xmin = FLOAT_TO_FIXED(tr.topX1)
	xmax = FLOAT_TO_FIXED(tr.topX2)
	
	for i=tr.topY to tr.bottomY
		if xmin=xmax then goto nextLine
		
		mincell = RoundDown(FIXED_TO_INT(xmin),16) shr 4
		maxcell = RoundUp  (FIXED_TO_INT(xmax),16) shr 4
		
		dstptr = texbuff+i*scr_pitch+(mincell shl 6)'*16*4
		
		for j=mincell to maxcell
			u = PerspU(j,i)
			v = PerspV(j,i)
			ustep = (PerspU(j+1,i)-u) shr 4
			vstep = (PerspV(j+1,i)-v) shr 4
			
			asm
				mov [backup_esp], esp
				mov [backup_ebp], ebp
				
				mov edx, [dstptr]
				
				mov esi, [u]
				mov edi, [v]
				mov esp, [ustep]
				mov ebp, [vstep]
				
				DrawTexTrap_Iter
				DrawTexTrap_Iter
				DrawTexTrap_Iter
				DrawTexTrap_Iter
				
				DrawTexTrap_Iter
				DrawTexTrap_Iter
				DrawTexTrap_Iter
				DrawTexTrap_Iter
				
				DrawTexTrap_Iter
				DrawTexTrap_Iter
				DrawTexTrap_Iter
				DrawTexTrap_Iter
				
				DrawTexTrap_Iter
				DrawTexTrap_Iter
				DrawTexTrap_Iter
				DrawTexTrap_Iter
				
				mov esp, [backup_esp]
				mov ebp, [backup_ebp]
				
				mov [dstptr], edx
			End Asm
		Next
		
		nextLine:
		xmin += xminstep
		xmax += xmaxstep
		
	Next
	
End Sub

#macro DrawColTrap_Iter
	#if 0
		asm
	#endif	
	'get uv
	mov ebx, esi
	mov eax, edi
	
	'dither (sloooow)
	'TODO: fix for 16bit fixed frac or even make adaptive
	mov ecx, [yf]
	shl ecx, 1
	or ecx, [xf]
	shl ecx, 8
	add eax, ecx
	sub eax, 512
	mov ecx, [xf]
	shl ecx, 9
	or ecx, 256
	mov edx, [yf]
	shl edx, 8
	add ebx, ecx
	sub ebx, edx
	sub edx, 512
	xor byte ptr [xf], 1
	
	shr ebx, FIXED_BITS
	shr eax, FIXED_BITS
	and ebx, [lightmap_maxx]
	and eax, [lightmap_maxy]
	
	'calc texel pointer
	mov cl, [lightmap_pitchlog]
	shl eax, cl 'a *= curtex_w
	shl ebx, 2
	add eax, ebx 'a += x
	add eax, [lightmap] 'a += curtex
	
	'read texel value
	mov eax, [eax]
	
	'draw pixel
	mov ebx, [dstptr]
	mov [ebx], eax
	
	'move pointer
	add dword ptr [dstptr], 4
	
	'update uv
	add esi, esp
	add edi, ebp
	
	#if 0
		end asm
	#endif
#EndMacro

sub DrawColTrap(tr as Trap)
	static as integer backup_esp, backup_ebp
	
	dim as integer i,j,k,fastmin,fastmax,mincell,maxcell
	dim as FIXED xmin,xmax,xminstep,xmaxstep
	
	static as integer xf, yf 'for dithering
	xf = 0
	yf = 0
	
	static as any ptr dstptr
	
	FixTrap(tr)
	
	if tr.bottomY=tr.topY then	return
	if tr.bottomY >= scr_h then tr.bottomY = scr_h-1
	if tr.topY < 0 then tr.topY = 0
	
	dim as FIXED u,v,ustep,vstep
	
	xminstep = FLOAT_TO_FIXED((tr.bottomX1-tr.topX1)/(tr.bottomY-tr.topY))
	xmaxstep = FLOAT_TO_FIXED((tr.bottomX2-tr.topX2)/(tr.bottomY-tr.topY))
	
	xmin = FLOAT_TO_FIXED(tr.topX1)
	xmax = FLOAT_TO_FIXED(tr.topX2)
	
	const OFFSETV as FIXED=FLOAT_TO_FIXED(+0.5)
	
	for i=tr.topY to tr.bottomY
		if xmin=xmax then goto nextLine
		
		'todo: maybe inline and optimize this?
		mincell = RoundDown(FIXED_TO_INT(xmin),16) shr 4
		maxcell = RoundUp  (FIXED_TO_INT(xmax),16) shr 4
		
		dstptr = colbuff+i*scr_pitch+(mincell shl 6)'*16*4
		
		yf = i and 1
		for j=mincell to maxcell
			u = PerspU(j,i)
			v = PerspV(j,i)+OFFSETV
			ustep = (PerspU(j+1,i)-u) shr 4
			vstep = (PerspV(j+1,i)+OFFSETV-v) shr 4
			
			asm
				mov [backup_esp], esp
				mov [backup_ebp], ebp
				
				'mov edx, [dstptr]
				
				mov esi, [u]
				mov edi, [v]
				mov esp, [ustep]
				mov ebp, [vstep]
				
				DrawColTrap_Iter
				DrawColTrap_Iter
				DrawColTrap_Iter
				DrawColTrap_Iter
				
				DrawColTrap_Iter
				DrawColTrap_Iter
				DrawColTrap_Iter
				DrawColTrap_Iter
				
				DrawColTrap_Iter
				DrawColTrap_Iter
				DrawColTrap_Iter
				DrawColTrap_Iter
				
				DrawColTrap_Iter
				DrawColTrap_Iter
				DrawColTrap_Iter
				DrawColTrap_Iter
				
				mov esp, [backup_esp]
				mov ebp, [backup_ebp]
				
				'mov [dstptr], edx
			End Asm
		Next
		
		nextLine:
		xmin += xminstep
		xmax += xmaxstep
		
	Next
	
End Sub

' Draw the low-res transformation grid (coords are in screen scale)
sub DrawPerspTriGrid(coords as Tri, UVs as Tri, depths as TriDepth)
	
	' Scale to grid space
	dim as single sx1,sy1,sx2,sy2,sx3,sy3
	sx1 = (coords.x1)/16
	sy1 = (coords.y1)
	sx2 = (coords.x2)/16
	sy2 = (coords.y2)
	sx3 = (coords.x3)/16
	sy3 = (coords.y3)
	
	dim as integer Xmin,Ymin,Xmax,Ymax
	XMin = min(sx1,sx2,sx3)-1
	YMin = min(sy1,sy2,sy3)-16
	XMax = max(sx1,sx2,sx3)+1
	YMax = max(sy1,sy2,sy3)+16
	
	if XMin<0 then Xmin=0
	if YMin<0 then Ymin=0
	if XMax>UBound(PerspU) then Xmax=UBound(PerspU)
	if YMax>=scr_h then Ymax=scr_h-1
	
	dim as single bary1,bary2,bary3,b1min,b2min,b3min,b1max,b2max,b3max,b1step,b2step,b3step
	dim as single baryT
	baryT = 1/((sY2-sY3)*(sX1-sX3)+(sX3-sX2)*(sY1-sY3))
	
	dim mult as single
	dim as single mult1, mult2, mult3, mult4, mult5, mult6
	
	dim as integer x, y, y1
	
	dim as single inv_width = 1/(xMax-xMin+1)
	
	mult1 = (uvs.x1/depths.z1)
	mult2 = (uvs.x2/depths.z2)
	mult3 = (uvs.x3/depths.z3)
	mult4 = (uvs.y1/depths.z1)
	mult5 = (uvs.y2/depths.z2)
	mult6 = (uvs.y3/depths.z3)
	
	for y=yMin to yMax step 16
		
		b1min = ((sY2-sY3)*(xMin-sX3)+(sX3-sX2)*(y-sY3))*baryT
		b2min = ((sY3-sY1)*(xMin-sX3)+(sX1-sX3)*(y-sY3))*baryT
		b3min = (1) - b1min - b2min
		
		b1max = ((sY2-sY3)*(xMax-sX3)+(sX3-sX2)*(y-sY3))*baryT
		b2max = ((sY3-sY1)*(xMax-sX3)+(sX1-sX3)*(y-sY3))*baryT
		b3max = (1) - b1max - b2max
		
		b1step = (b1max-b1min)*inv_width
		b2step = (b2max-b2min)*inv_width
		b3step = (b3max-b3min)*inv_width
		
		bary1 = b1min
		bary2 = b2min
		bary3 = b3min
		
		for x=xMin to xMax
			
			mult = 1/(bary1/depths.z1 + bary2/depths.z2 + bary3/depths.z3)
			
			PerspU(x,y) = FLOAT_TO_FIXED((bary1*mult1 + bary2*mult2 + bary3*mult3)*mult)
			PerspV(x,y) = FLOAT_TO_FIXED((bary1*mult4 + bary2*mult5 + bary3*mult6)*mult)
			
			'PerspZ(x,y) = bary1*z1 + bary2*z2 + bary3*z3
			
			bary1 += b1step
			bary2 += b2step
			bary3 += b3step
			
		Next
	Next
	
	dim as FIXED u, v, ustep, vstep
	
	' interpolate
	for y=yMin to yMax step 16
		for x=xMin to xMax
			u = PerspU(x,y)
			v = PerspV(x,y)
			ustep = (PerspU(x,y+16)-u) shr 4
			vstep = (PerspV(x,y+16)-v) shr 4
			u += ustep
			v += vstep
			for y1=y to y+15
				PerspU(x,y1) = u
				PerspV(x,y1) = v
				u += ustep
				v += vstep
			Next
		Next
	Next
	
End Sub

/'
	trafgrid = 9.5  ms
	buffcopy = 5.2  ms
	drawtrap = 17.8 ms
'/

sub DrawPerspTri(coords as Tri, UVs as Tri, LightUVs as Tri, depths as TriDepth) export
	dim as Trap top, bottom
	
	SplitTri(coords,top,bottom)
	DrawPerspTriGrid(coords,UVs,depths)
	DrawTexTrap(top)
	DrawTexTrap(bottom)
	
	if lightmap<>0 then
		DrawPerspTriGrid(coords,LightUVs,depths)
		DrawColTrap(top)
		DrawColTrap(bottom)
		BuffCopy(texbuff,colbuff,top   ,Modulated)
		BuffCopy(texbuff,colbuff,bottom,Modulated)
	EndIf
	
	BuffCopy(scr,texbuff,top   ,curtex_style)
	BuffCopy(scr,texbuff,bottom,curtex_style)
End Sub
